"""
=====================CAUTION=======================
DO NOT DELETE THIS FILE SINCE IT IS PART OF NROBO
FRAMEWORK AND IT MAY CHANGE IN THE FUTURE UPGRADES
OF NROBO FRAMEWORK. THUS, TO BE ABLE TO SAFELY UPGRADE
TO LATEST NROBO VERSION, PLEASE DO NOT DELETE THIS
FILE OR ALTER ITS LOCATION OR ALTER ITS CONTENT!!!
===================================================

@author: Panchdev Singh Chauhan
@email: erpanchdev@gmail.com
"""

import os

from nrobo import *
from nrobo.exceptions import *
from nrobo.util.commands.posix import POSIX_COMMAND
from nrobo.util.commands.windows import WINDOWS_COMMAND
from nrobo.util.platform import PLATFORMS
from nrobo.util.process import terminal, terminal_nogui


class N_COMMANDS:
    """N_COMMANDS class hold nrobo command mappings for host platform.

    Raises <MissingCommandImplementation> exception
    If there is no implementation for host platform."""

    CLEAR_SCREEN = "clear screen"
    COMMAND = {
        PLATFORMS.WINDOWS: {
            CLEAR_SCREEN: WINDOWS_COMMAND.CLS,
        },
        PLATFORMS.DARWIN: {
            CLEAR_SCREEN: POSIX_COMMAND.CLEAR
        }
    }


def get_command(command) -> None:
    """Return the appropriate posix or windows <command> to caller."""

    try:
        return N_COMMANDS.COMMAND[os.environ[EnvKeys.HOST_PLATFORM]][N_COMMANDS.CLEAR_SCREEN]
    except KeyError as ke:
        raise MissingCommandImplementation(N_COMMANDS.CLEAR_SCREEN)


def clear_screen():
    """Run the clear screen command."""
    terminal([get_command(N_COMMANDS.CLEAR_SCREEN)])


def remove_files_recursively(directory) -> int:
    """Remove <directory>

    :param directory:
    :return:"""

    if os.environ[EnvKeys.HOST_PLATFORM] in [PLATFORMS.DARWIN, PLATFORMS.LINUX, PLATFORMS.MACOS]:
        try:
            return terminal(["rm", "-rf", directory])
        except Exception as e:
            print(e)

    if os.environ[EnvKeys.HOST_PLATFORM] in [PLATFORMS.WINDOWS]:
        try:
            return terminal(["del", "/q", "/S", directory + os.sep + "*.*"])
        except Exception as e:
            print(e)

    return 1



"""
=====================CAUTION=======================
DO NOT DELETE THIS FILE SINCE IT IS PART OF NROBO
FRAMEWORK AND IT MAY CHANGE IN THE FUTURE UPGRADES
OF NROBO FRAMEWORK. THUS, TO BE ABLE TO SAFELY UPGRADE
TO LATEST NROBO VERSION, PLEASE DO NOT DELETE THIS
FILE OR ALTER ITS LOCATION OR ALTER ITS CONTENT!!!
===================================================

@author: Panchdev Singh Chauhan
@email: erpanchdev@gmail.com
"""

import json
import os.path as path
import string
import sys
import random
from pathlib import Path
from time import time
from typing import Union

import yaml


class Common:
    """Customized Selenium WebDriver class which contains all the useful methods that can be re used.
    These methods _help to in the following cases:
    To reduce the time required to write automation script.
    To take the screenshot in case of test case failure.
    To log to provide waits"""

    @staticmethod
    def read_file_as_string(file_path: Union[str, Path], encoding=None):
        """
        Read file as string

        :param file_path:
        :param encoding:
        :return:
        """

        try:
            if encoding is None:
                with open(file_path, "r") as f:
                    content = f.read()
                    return content
            else:
                with open(file_path, "r", encoding=encoding) as f:
                    content = f.read()
                    return content
        except FileNotFoundError as file_not_found_error:
            print("No such file or directory found: " + file_path)
            raise file_not_found_error

    @staticmethod
    def write_text_to_file(file_path: Union[str, Path], content, encoding=None):
        """
        Write text to file

        :param file_path:
        :param content:
        :param encoding:
        :return:
        """
        if isinstance(file_path, str):
            file_path = Path(file_path)

        try:
            if encoding is None:
                with open(file_path, 'w') as f:
                    f.write(content)
            else:
                with open(file_path, 'w', encoding=encoding) as f:
                    f.write(content)

        except FileNotFoundError as file_not_found_error:
            print("No such file or directory found: " + str(file_path))

    @staticmethod
    def append_text_to_file(file_path: Union[str, Path], content, encoding=None):
        """
        Write text to file

        :param file_path:
        :param content:
        :param encoding:
        :return:
        """
        if isinstance(file_path, str):
            file_path = Path(file_path)

        try:
            if encoding is None:
                with open(file_path, 'a') as f:
                    f.write(content)
            else:
                with open(file_path, 'w', encoding=encoding) as f:
                    f.write(content)

        except FileNotFoundError as file_not_found_error:
            print("No such file or directory found: " + str(file_path))

    @staticmethod
    def read_json(file_path: Union[str, Path]):
        """
        Read Json

        :param file_path:
        :return:
        """

        try:

            with open(file_path) as f:
                data = json.load(f)
                return data

        except FileNotFoundError as file_not_found_error:
            print("No such file or directory found: " + file_path)

    @staticmethod
    def write_json(file_path: Union[str, Path], dictionary):
        """
        Write dictionary data to file

        :param file_path: Path of file where dictionary date is going to be stored
        :param dictionary: Dictionary of data
        :return:
        """

        with open(file_path, 'w') as file:  # Open given file in write mode
            json.dump(dictionary, file, sort_keys=True, indent=4)

    @staticmethod
    def is_file_exist(file_path: Union[str, Path]):
        """
        Checks if given file exists or not.

        :param file_path: Path of file
        :return: True if file exists else return False
        """
        return path.exists(file_path)

    @staticmethod
    def read_yaml(file_path: Union[str, Path], /, *, fail_on_failure=True) -> Union[str, None]:
        """
        Read yaml file at given path

        :param fail_on_failure: create file if TRUE though file is not present_release to read.
        :param file_path: Path of file
        :return: Content of yaml file -> dict()
        """

        if not path.exists(file_path) and not fail_on_failure:
            """if file does not exist, then let's create it first"""

            with open(file_path, 'w') as file:
                """Create a file"""

                # initialize file with empty dictionary
                yaml.dump({}, file)

        if not path.exists(str(file_path)) and fail_on_failure:
            """file does not exist"""
            raise Exception(f"File {file_path} does not exist!")
        else:
            """Do Nothing as file exists"""
            pass

        # Read the file
        with open(r'{0}'.format(file_path)) as file:
            # The FullLoader parameter handles the conversion from YAML
            # scalar values to Python the dictionary format
            data = yaml.load(file, Loader=yaml.SafeLoader)

            # return with data as dictionary
            return data

    @staticmethod
    def write_yaml(file_path: Union[str, Path], dictionary):
        """
        Write dictionary data to given file path in yaml format

        :param file_path: Path of file where dictionary data needs to be stored
        :param dictionary: Dictionary data
        :return: Nothin
        """

        with open(file_path, 'w') as file:  # Open given file in write mode
            yaml.dump(dictionary, file)

    @staticmethod
    def generate_random_numbers(min, max):
        """
        Generate and return a random number in given range denoted by min and max

        :param min:
        :param max:
        :return:
        """
        """
        Returns a random string of given length

        @Returns string a random string
        """
        random_number = random.randint(min, max)

        return random_number

    @staticmethod
    def save_as_pdf(file_content_as_string, path: [str, Path] = None):
        """Save as pdf to given path"""

        import base64

        if path is None:
            path = "downloads/temp.pdf"

        if isinstance(path, Path):
            path = str(path)

        with open(path, 'wb') as theFile:
            theFile.write(base64.b64decode(file_content_as_string))

    @staticmethod
    def save_base64string(file_content_as_string, path: [str, Path] = None):
        """Save base64string as given path"""

        import base64

        if path is None:
            path = "downloads/temp.pdf"

        if isinstance(path, Path):
            path = str(path)

        with open(path, 'wb') as theFile:
            theFile.write(base64.b64decode(file_content_as_string))

    @staticmethod
    def save_bytes_to_file(bytes, path: [str, Path] = None):
        """Save bytes to given path"""

        import base64

        if path is None:
            path = "temp.jpg"

        if isinstance(path, Path):
            path = str(path)

        with open(path, 'wb') as theFile:
            theFile.write(bytes)



"""
=====================CAUTION=======================
DO NOT DELETE THIS FILE SINCE IT IS PART OF NROBO
FRAMEWORK AND IT MAY CHANGE IN THE FUTURE UPGRADES
OF NROBO FRAMEWORK. THUS, TO BE ABLE TO SAFELY UPGRADE
TO LATEST NROBO VERSION, PLEASE DO NOT DELETE THIS
FILE OR ALTER ITS LOCATION OR ALTER ITS CONTENT!!!
===================================================


Definitions of CONSTANTS.

@author: Panchdev Singh Chauhan
@email: erpanchdev@gmail.com
"""


class CONST(object):
    """
    Collection of CONSTANTS
    """
    NEWLINE = "\n"

    SINGLE_INVERTED_COMMA = "'"
    EMPTY = ""
    EQUAL = "="
    BLANK = EMPTY
    SPACE = " "
    STAR = "*"
    ASTERISK = STAR
    UNDERSCORE = "_"
    FORWARD_SLASH = "/"
    MINUS = "-"
    HYPHEN = MINUS
    TILD = "~"
    CARET = "`"
    QUESTION = "?"
    EXCLAMATION = "!"
    HASH = "#"

    # Brackets & PROGRAMING LANGUAGES
    PARENTHESIS_OPEN = "("
    PARENTHESIS_CLOSE = ")"
    CURLY_BRACE_OPEN = "{"
    CURLY_BRACE_CLOSE = "}"
    SQUARE_BRACKET_OPEN = "["
    SQUARE_BRACKET_CLOSE = "]"
    SCOPE_RESOLUTION_OPERATOR = "::"

    # Web and browser
    HTTPS = "https://"
    HTTP = "http://"
    DOT = "."
    COLON = ":"
    ADDRESS_OF = "@"
    AT_THE_RATE = ADDRESS_OF

    # Day, Date and Time
    HOURS_PER_DAY = 24
    MINUTES_PER_HOUR = 60
    SECONDS_PER_MINUTE = 60
    PERCENTAGE = '%'

    # Currencies
    DOLLAR = "$"
    RUPEE = u'\u20B9'

    # Greetings
    FOLDED_HAND = '\uF09F\u998F'
    HEART_RED = '\u2764\ufe0f'

    # Path seperators
    SLASH = FORWARD_SLASH
    BACKSLASH = u'\u005C'


class EXT:
    """Constants for various extensions."""

    YAML = ".yaml"
    PY = ".py"
    PYC = ".pyc"



"""
=====================CAUTION=======================
DO NOT DELETE THIS FILE SINCE IT IS PART OF NROBO
FRAMEWORK AND IT MAY CHANGE IN THE FUTURE UPGRADES
OF NROBO FRAMEWORK. THUS, TO BE ABLE TO SAFELY UPGRADE
TO LATEST NROBO VERSION, PLEASE DO NOT DELETE THIS
FILE OR ALTER ITS LOCATION OR ALTER ITS CONTENT!!!
===================================================

Trigger for nrobo framework!

@author: Panchdev Singh Chauhan
@email: erpanchdev@gmail.com

"""


def main():
    """Entry point of nrobo command-line-utility."""

    try:
        import os
        from nrobo.util.process import terminal
        import subprocess
        terminal(["pip", "install", "PyYAML"], stdout=subprocess.DEVNULL, stderr=subprocess.STDOUT)
        from nrobo.cli.launcher import launch_nrobo, launcher_command
        from nrobo.cli.upgrade import confirm_update
        from nrobo import EnvKeys, NROBO_CONST, NROBO_PATHS
        from nrobo import greet_the_guest, NROBO_CONST, EnvKeys
        from nrobo.cli.nrobo_args import nrobo_cli_parser
        from nrobo.cli.install import install_nrobo
        from nrobo.util.commands.ncommands import clear_screen, remove_files_recursively
        from nrobo.util.process import terminal
        from nrobo.util.constants import CONST
        from nrobo.cli.cli_constants import NREPORT
        from nrobo.util.python import verify_set_python_install_pip_command

        # clear screen
        clear_screen()

        # called to set EnvKeys dependent on args
        command, args, command_builder_notes = launcher_command()
        if command is None:
            exit(0)

        # greet the guest
        greet_the_guest()

        # install dependencies
        install_nrobo(install_only=True)

        # verify python installation and version
        verify_set_python_install_pip_command()

        # remove 'dist' directory created by python build module during packaging
        remove_files_recursively(NROBO_CONST.DIST_DIR)

        # delete results directory created by nrobo for storing test results
        remove_files_recursively(NREPORT.REPORT_DIR)

        # parse nrobo cli arguments
        launch_nrobo()

    except KeyboardInterrupt as e:
        print(e)

    return 0



"""
=====================CAUTION=======================
DO NOT DELETE THIS FILE SINCE IT IS PART OF NROBO
FRAMEWORK AND IT MAY CHANGE IN THE FUTURE UPGRADES
OF NROBO FRAMEWORK. THUS, TO BE ABLE TO SAFELY UPGRADE
TO LATEST NROBO VERSION, PLEASE DO NOT DELETE THIS
FILE OR ALTER ITS LOCATION OR ALTER ITS CONTENT!!!
===================================================

nRoBo Constants.

@author: Panchdev Singh Chauhan
@email: erpanchdev@gmail.com
"""
from nrobo.exceptions import NRoBoBrowserNotSupported

from rich.console import *
from nrobo.cli.formatting import *

from nrobo.cli import *
console = Console(theme=themes)


class Browsers:
    """browser names constants"""

    CHROME = 'chrome'
    CHROME_HEADLESS = "chrome_headless"
    ANTI_BOT_CHROME = "anti_bot_chrome"
    EDGE = 'edge'
    FIREFOX = 'firefox'
    FIREFOX_HEADLESS = "firefox_headless"
    IE = "ie"
    SAFARI = 'safari'
    OPERA = 'opera'


# list holding supported browser in nRoBo framework
supported_browsers = [Browsers.CHROME, Browsers.CHROME_HEADLESS, Browsers.ANTI_BOT_CHROME,
                      Browsers.SAFARI,
                      Browsers.FIREFOX, Browsers.FIREFOX_HEADLESS, Browsers.EDGE, Browsers.IE]
# # list holding browsers not supported in nRoBo framework
supported_browsers_in_future = [Browsers.OPERA]


def raise_exception_if_browser_not_supported(browser_name):
    """raise exception if <browser_name> is not supported in nrobo framework"""

    if str(browser_name).lower() in supported_browsers_in_future:
        console.print(
            f"[{STYLE.HLOrange}]Support for {browser_name} browser will be coming in upcoming releases. Sorry for the "
            f"convenience.")
        exit(1)
    if browser_name not in supported_browsers:
        raise NRoBoBrowserNotSupported(browser_name)







"""
@author: Panchdev Singh Chauhan
@email: erpanchdev@gmail.com
"""

import logging

from selenium.webdriver.common.by import By
from pages import Page
from pages.PageSearch import PageSearch

class PagePyPiHome(Page):
    """Page class for PyPi.org home page"""

    def __init__(self, driver, logger):
        """page constructor"""

        # Mandatory call to parent constructor
        # Always place as a first statement in each of your Page Classes
        super().__init__(driver, logger)

        # other initialization may proceed after this

    # Page locators definition should go here
    # ----------------------------------------
    url = "https://pypi.org/"
    txt_search = (By.ID, 'search')
    btn_search = (By.CSS_SELECTOR, ".search-form__button")
    # lnkLogin = (By.ID, 'login2')
    # btnSubmit = (By, Value)
    # txtName = (By, Value)
    # ...

    # Page method specification should go here
    # ----------------------------------------

    def open(self):
        self.logger.info(f"Open url: {self.url}")
        self.get(self.url)

    def search_button_present(self):
        return self.find_element(*self.txt_search).is_displayed()

    def type_search_keyword(self, keyword):
        self.logger.info(f"Type search keyword: {keyword}")
        self.send_keys(*self.txt_search, keyword)

    def search(self) -> PageSearch:
        self.logger.info(f"Click on Search button.")
        self.click(*self.btn_search)

        return PageSearch(self.driver, self.logger)



"""
=====================CAUTION=======================
DO NOT DELETE THIS FILE SINCE IT IS PART OF NROBO
FRAMEWORK AND IT MAY CHANGE IN THE FUTURE UPGRADES
OF NROBO FRAMEWORK. THUS, TO BE ABLE TO SAFELY UPGRADE
TO LATEST NROBO VERSION, PLEASE DO NOT DELETE THIS
FILE OR ALTER ITS LOCATION OR ALTER ITS CONTENT!!!
===================================================

Version class handles version operations smoothly.


@author: Panchdev Singh Chauhan
@email: erpanchdev@gmail.com
"""
from nrobo.exceptions import NRoBoIncorrectVersion, NRoBoInvalidOperation


class Version:
    def __init__(self, version):
        self._version = version
        self._major = int(self._version_parts()[0])
        self._minor = int(self._version_parts()[1])
        self._patch = int(self._version_parts()[2])
        self._major_next = self._major + 1
        self._major_prev = self._major - 1
        self._minor_next = self._minor + 1
        self._minor_prev = self._minor - 1
        self._patch_next = self._patch + 1
        self._patch_prev = self._patch - 1
        self._version_next = f"{self._major}.{self._minor}.{self._patch_next}"
        self._version_prev = f"{self._major}.{self._minor}.{self._patch_prev}"

    @property
    def version(self):
        return self._version

    @version.setter
    def version(self, version):
        self._version = version

    @property
    def version_next(self):
        return self._version_next

    @version_next.setter
    def version_next(self, version_next):
        self._version_next = version_next

    @property
    def version_prev(self):
        return self._version_prev

    @version_prev.setter
    def version_prev(self, version_prev):
        self._version_prev = version_prev

    @property
    def major(self):
        return self._major

    @major.setter
    def major(self, major):
        self._major = major

    @property
    def major_next(self):
        return self._major_next

    @major_next.setter
    def major_next(self, major_next):
        self._major_next = major_next

    @property
    def major_prev(self):
        return self._major_prev

    @major_prev.setter
    def major_prev(self, major_prev):
        self._major_prev = major_prev

    @property
    def minor(self):
        return self._minor

    @minor.setter
    def minor(self, minor):
        self._minor = minor

    @property
    def minor_next(self):
        return self._minor_next

    @minor_next.setter
    def minor_next(self, minor_next):
        self._minor_next = minor_next

    @property
    def minor_prev(self):
        return self._minor_prev

    @minor_prev.setter
    def minor_prev(self, minor_prev):
        self._minor_prev = minor_prev

    @property
    def patch(self):
        return self._patch

    @patch.setter
    def patch(self, patch):
        self._patch = patch

    @property
    def patch_next(self):
        return self._patch_next

    @patch_next.setter
    def patch_next(self, patch_next):
        self._patch_next = patch_next

    @property
    def patch_prev(self):
        return self._patch_prev

    @patch_prev.setter
    def patch_prev(self, patch_prev):
        self._patch_prev = patch_prev

    def major_incremented(self) -> str:
        return f"{self._major_next}.0.0"

    def major_decremented(self) -> str:
        return f"{self._major_prev}.0.0"

    def minor_incremented(self) -> str:
        return f"{self._major}.{self._minor_next}.0"

    def minor_decremented(self) -> str:
        return f"{self._major}.{self._minor_prev}.0"

    def patch_incremented(self) -> str:
        return f"{self._major}.{self._minor}.{self._patch_next}"

    def patch_decremented(self) -> str:
        return f"{self._major}.{self._minor}.{self._patch_prev}"

    def version_incremented(self) -> str:
        return f"{self._major}.{self._minor}.{self._patch_next}"

    def version_decremented(self) -> str:
        return f"{self._major}.{self._minor}.{self._patch_prev}"

    @staticmethod
    def first_major_release(release_version):
        _release_version = Version(release_version)
        return f"{_release_version._major}.0.0"

    @staticmethod
    def first_minor_release(release_version):
        _release_version = Version(release_version)
        return f"{_release_version._major}.{_release_version._minor}.{0}"

    @staticmethod
    def present_is_a_major_release(present_release: str, previous_release: str) -> bool:
        """Returns True if present release is a major release.

        :param present_release:
        :param previous_release:
        :return: """
        _present_release = Version(present_release)
        _previous_release = Version(previous_release)

        if _present_release.major > _previous_release.major:
            return True

        return False

    @staticmethod
    def present_is_a_minor_release(present_release: str, previous_release: str) -> bool:
        """Returns True if present release is a minor release.

        :param present_release:
        :param previous_release:
        :return: """
        _present_release = Version(present_release)
        _previous_release = Version(previous_release)

        if _present_release.minor > _previous_release.minor:
            return True

        return False

    @staticmethod
    def present_is_a_patch_release(present_release: str, previous_release: str) -> bool:
        """Returns True if present release is a patch release.

        :param present_release:
        :param previous_release:
        :return: """
        _present_release = Version(present_release)
        _previous_release = Version(previous_release)

        if _present_release.patch > _previous_release.patch:
            return True

        return False

    def _version_parts(self):
        patten = r"(([\d]*)[.]([\d]*)[.]([\d]*))"
        import re
        m = re.match(patten, self._version)

        if not m:
            # incorrect version
            raise NRoBoIncorrectVersion(self._version)

        _version = m.group(1)
        _major = m.group(2)
        _minor = m.group(3)
        _patch = m.group(4)

        return [_major, _minor, _patch]

    def __lt__(self, other):
        if not isinstance(other, Version):
            raise NRoBoInvalidOperation("<", type(other))

        if self.major < other.major:
            return True

        if self.minor < other.minor:
            return True

        if self.patch < other.patch:
            return True

        return False

    def __eq__(self, other):
        if not isinstance(other, Version):
            raise NRoBoInvalidOperation("=", type(other))

        if self.major == other.major \
                and self.minor == other.minor \
                and self.patch == other.patch:
            return True

        return False

    def __gt__(self, other):
        if not isinstance(other, Version):
            raise NRoBoInvalidOperation(">", type(other))

        if self.major > other.major:
            return True

        if self.minor > other.minor:
            return True

        if self.patch > other.patch:
            return True

        return False

    def __add__(self, other):
        if not isinstance(other, int) \
                and not isinstance(other, float):
            raise NRoBoInvalidOperation("+", type(other))

        if isinstance(other, float):
            other = int(float)

        return Version(f"{self.major}.{self.minor}.{self.patch + other}")

    def __sub__(self, other):
        if not isinstance(other, int) \
                and not isinstance(other, float):
            raise NRoBoInvalidOperation("-", type(other))

        if isinstance(other, float):
            other = int(float)

        return Version(f"{self.major}.{self.minor}.{self.patch - other}")

    def __le__(self, other):
        if not isinstance(other, Version):
            raise NRoBoInvalidOperation("<=", type(other))

        if self.__lt__(other):
            return True

        if self.__eq__(other):
            return True

        return False

    def __ge__(self, other):
        if not isinstance(other, Version):
            raise NRoBoInvalidOperation(">=", type(other))

        if self.__gt__(other):
            return True

        if self.__eq__(other):
            return True

        return False



"""
=====================CAUTION=======================
DO NOT DELETE THIS FILE SINCE IT IS PART OF NROBO
FRAMEWORK AND IT MAY CHANGE IN THE FUTURE UPGRADES
OF NROBO FRAMEWORK. THUS, TO BE ABLE TO SAFELY UPGRADE
TO LATEST NROBO VERSION, PLEASE DO NOT DELETE THIS
FILE OR ALTER ITS LOCATION OR ALTER ITS CONTENT!!!
===================================================


nrobo module loads nRoBo globals.


@author: Panchdev Singh Chauhan
@email: erpanchdev@gmail.com
"""

__version__ = '2024.27.0'

# install rich library
import os
from pathlib import Path


class NROBO_CONST:
    """nrobo special constants"""

    NROBO = "nrobo"
    DIST_DIR = "dist"
    SUCCESS = 0


class Python:
    """Information related to python"""

    PIP = "pip"
    VERSION = "Version"
    PYPINFO = "pypinfo"


class Environment:
    """Environments"""

    PRODUCTION = "Production"
    DEVELOPMENT = "Development"


class EnvKeys:
    """nRoBo environment keys

    Example:
        PIP_COMMAND = pip | pip3

        and many more such...
    """
    APPIUM = "appium"
    PIP_COMMAND = "Pip Command"
    EXEC_DIR = "Execution Directory"
    NROBO_DIR = "nRoBo Installation Directory"
    ENVIRONMENT = "Environment"
    PYTHON = "Python"
    APP = "App Name"
    URL = "App Url"
    USERNAME = "Username"
    PASSWORD = "Password"
    TITLE = "Title"
    BROWSER = "Browser"
    HOST_PLATFORM = "Host Platform"
    DEBUG = "debug mode"
    SUPPRESS_PROMPT = "suppress prompt"


# load environment keys with defaults
os.environ[EnvKeys.APPIUM] = "0"
os.environ[EnvKeys.PIP_COMMAND] = Python.PIP
os.environ[EnvKeys.EXEC_DIR] = ""
os.environ[EnvKeys.NROBO_DIR] = ""
os.environ[EnvKeys.ENVIRONMENT] = Environment.PRODUCTION
os.environ[EnvKeys.PYTHON] = "python"
os.environ[EnvKeys.APP] = "nRoBo"
os.environ[EnvKeys.URL] = ""
os.environ[EnvKeys.USERNAME] = ""
os.environ[EnvKeys.PASSWORD] = ""
os.environ[EnvKeys.TITLE] = ""
os.environ[EnvKeys.BROWSER] = ""
os.environ[EnvKeys.HOST_PLATFORM] = ""
os.environ[EnvKeys.DEBUG] = "False"
os.environ[EnvKeys.SUPPRESS_PROMPT] = "0"


class NROBO_PATHS:
    """nRoBo framework directories and files"""
    EXEC_DIR = Path(os.environ[EnvKeys.EXEC_DIR])
    NROBO_DIR = Path(os.environ[EnvKeys.NROBO_DIR])
    NODE_MODULES = Path("node_modules")
    PATCHES = Path("patches")
    NROBO = Path("nrobo")
    INIT_PY = Path("__init__.py")
    BROWSER_CONFIGS = Path("browserConfigs")
    CAPABILITY_YAML = BROWSER_CONFIGS / "capability.yaml"
    CAPABILITY_APPIUM_YAML = BROWSER_CONFIGS / "appium_capability.yaml"
    MARKERS_YAML = BROWSER_CONFIGS / "markers.yaml"

    # Browsers packages
    BROWSERS = Path("browsers")
    CHROME = Path("chrome")
    BROWSERS_CHROME_PKG = BROWSERS / CHROME
    EDGE = Path("edge")
    BROWSERS_EDGE_PKG = BROWSERS / EDGE
    FIREFOX = Path("firefox")
    BROWSERS_FIREFOX_PKG = BROWSERS / FIREFOX
    SAFARI = Path("safari")
    BROWSERS_SAFARI_PKG = BROWSERS / SAFARI

    # cli packages
    CLI = Path("cli")
    CLI_PKG = CLI / INIT_PY
    DETECTION = Path("detection")
    DETECTION_PKG = CLI / DETECTION / INIT_PY
    FORMATTING = Path("formatting")
    FORMATTING_PKG = CLI / FORMATTING / INIT_PY
    INSTALL = Path("install")
    INSTALL_PKG = CLI / INSTALL / INIT_PY
    NCODES = Path("ncodes")
    NCODES_PKG = CLI / NCODES / INIT_PY
    REQUIREMENTS_TXT_FILE = CLI / INSTALL / Path("requirements.txt")
    CLI_NROBO_ARGS = Path("nrobo_args")
    CLI_NROBO_ARGS_PKG = CLI / CLI_NROBO_ARGS / INIT_PY
    CLI_TOOLS = Path("tools")
    CLI_TOOLS_PKG = CLI / CLI_TOOLS / INIT_PY
    UPGRADE = Path("upgrade")
    UPGRADE_PKG = CLI / UPGRADE / INIT_PY
    CLI_LAUNCHER_PY_FILE = CLI / Path("launcher.py")
    CLI_CONSTANTS_PY_FILE = CLI / Path("cli_constants.py")
    CLI_VERSION_YAML_FILE = CLI / Path("cli_version.yaml")
    NGLOBALS_PY_FILE = CLI / Path("nglobals.py")

    EXCEPTIONS = Path("exceptions")

    # framework packages
    FRAMEWORK = Path("framework")
    PAGES = Path("pages")
    FRAMEWORK_PAGES = FRAMEWORK / PAGES
    FRAMEWORK_PAGE_PYPI_HOME_PY_FILE = FRAMEWORK_PAGES / Path("PagePyPiHome.py")
    TESTS = Path("tests")
    MOBILE = Path("mobile")
    WEB = Path("web")
    FRAMEWORK_TESTS = FRAMEWORK / TESTS
    GUI = FRAMEWORK_TESTS / WEB / Path("gui")
    GUI_PKG = GUI / INIT_PY
    GUI_PYPI_HOME_PAGE_TEST_PY_FILE = GUI / Path("PyPi_home_page_test.py")
    NO_GUI_PKG = TESTS / WEB / "no_gui" / INIT_PY

    FRAMEWORK_TESTS = FRAMEWORK / TESTS
    NROBO_CONFIG_FILE = Path("nrobo-config.yaml")
    FRAMEWORK_NROBO_CONFIG = FRAMEWORK / NROBO_CONFIG_FILE
    NROBO_CONFTEST_HOST_FILE = FRAMEWORK / "conftest-host.py"

    SELENESE = Path("selenese")

    # nrobo.util packages
    UTIL = Path("util")
    COMMANDS = Path("commands")
    NCOMMANDS = Path("ncommands")
    POSIX = Path("posix")
    WINDOWS = Path("windows")
    UTIL_NCOMMANDS_PKG = NROBO / UTIL / COMMANDS / NCOMMANDS / INIT_PY
    UTIL_POSIX_PKG = NROBO / UTIL / COMMANDS / POSIX / INIT_PY
    UTIL_WINDOWS_PKG = NROBO / UTIL / COMMANDS / WINDOWS / INIT_PY
    UTIL_COMMON = Path("common")
    UTIL_COMMON_PKG = NROBO / UTIL / UTIL_COMMON / INIT_PY
    UTIL_CONSTANT = Path("constants")
    UTIL_CONSTANT_PKG = NROBO / UTIL / UTIL_CONSTANT / INIT_PY
    UTIL_FILESYSTEM = Path("filesystem")
    UTIL_FILESYSTEM_PKG = NROBO / UTIL / UTIL_FILESYSTEM / INIT_PY
    UTIL_PLATFORM = Path("platform")
    UTIL_NETWORK = Path("network")
    UTIL_NETWORK_PKG = NROBO / UTIL / UTIL_NETWORK / INIT_PY
    UTIL_PLATFORM_PKG = NROBO / UTIL / UTIL_PLATFORM / INIT_PY
    UTIL_PROCESS = Path("process")
    UTIL_PROCESS_PKG = NROBO / UTIL / UTIL_PROCESS / INIT_PY
    UTIL_PYTHON = Path("python")
    UTIL_PYTHON_PKG = NROBO / UTIL / UTIL_PYTHON / INIT_PY
    UTIL_REGEX = Path("regex")
    UTIL_REGEX_PKG = NROBO / UTIL / UTIL_REGEX / INIT_PY
    UTIL_VERSION = Path("version")
    UTIL_VERSION_PKG = NROBO / UTIL / UTIL_VERSION / INIT_PY
    CONFTEST_PY = Path("conftest.py")

    NROBO_FRAMEWORK_TESTS = Path("framework_tests")
    TESTS = Path("tests")

    VERSIONS = Path("versions")
    PROD_YAML = VERSIONS / Path("prod.yaml")
    TEST_YAML = VERSIONS / Path("test.yaml")

    NROBO_PY_FILE = Path("nrobo.py")
    PACKAGE_PY_FILE = Path("package.py")
    LICENSE_FILE = Path("LICENSE")
    PY_PROJECT_TOML_FILE = Path("pyproject.toml")
    README_RST_FILE = Path("README.rst")
    VALIDATE_NROBO_PY_FILE = Path("validatenrobo.py")

    PYTEST_LIFE_CYCLE_LOGS = Path("key") / "experiments" / "pytest" / "pytest-life-cycle-logs"


class NROBO_CLI_TOOL_PATH:
    """nRoBo CLI tool paths"""
    CLI = Path("cli")
    BUILD = CLI / Path("build")
    CHECK = CLI / Path("check")
    COMPILE = CLI / Path("compile")
    DEVELOPMENT = CLI / Path("development")
    PUBLISH = CLI / Path("publish")


class NROBO_FRAMEWORK_TESTS:
    """nrobo_framework_tests package"""
    NROBO_FRAMEWORK_TESTS_CONFTEST_PY_FILE = NROBO_PATHS.NROBO_FRAMEWORK_TESTS / NROBO_PATHS.CONFTEST_PY
    TEST_NROBO_FRAMEWORK_PY_FILE = NROBO_PATHS.NROBO_FRAMEWORK_TESTS / Path("test_package_presence.py")


import subprocess
from nrobo.util.process import terminal

terminal(["pip", "install", "rich"], stdout=subprocess.DEVNULL, stderr=subprocess.STDOUT)

from pathlib import Path
import re
from rich.console import Console
from nrobo.cli.formatting import themes as th, STYLE
from nrobo.util.constants import CONST

# rich console
console = Console(theme=th)


def greet_the_guest():
    """greet the guest with Indian way of greeting!"""

    greet_msg = 'Namastey World!. Thank you for choosing, nRoBo.'
    formatted_heart_string = CONST.HEART_RED * len(greet_msg)

    console.print(f'[{STYLE.HLRed}]{formatted_heart_string}')
    console.print(f"[{STYLE.HLOrange}]{greet_msg}[/]")
    console.print(f'[{STYLE.HLRed}]{formatted_heart_string}')


def set_environment() -> None:
    """set nrobo environment

        Not complete implementation as the name suggests.
        This implementation will be corrected in future versions..."""

    # get directory from where the script was executed
    os.environ[EnvKeys.EXEC_DIR] = os.getcwd()
    # get directory where this script resides
    nrobo_loader_file_path = os.path.dirname(os.path.realpath(__file__))
    # grab nrobo installation path
    os.environ[EnvKeys.NROBO_DIR] = re.findall(f"(.*{NROBO_CONST.NROBO})", str(nrobo_loader_file_path))[0]

    NROBO_PATHS.EXEC_DIR = Path(os.environ[EnvKeys.EXEC_DIR])
    NROBO_PATHS.NROBO_DIR = Path(os.environ[EnvKeys.NROBO_DIR])



"""
=====================CAUTION=======================
DO NOT DELETE THIS FILE SINCE IT IS PART OF NROBO
FRAMEWORK AND IT MAY CHANGE IN THE FUTURE UPGRADES
OF NROBO FRAMEWORK. THUS, TO BE ABLE TO SAFELY UPGRADE
TO LATEST NROBO VERSION, PLEASE DO NOT DELETE THIS
FILE OR ALTER ITS LOCATION OR ALTER ITS CONTENT!!!
===================================================


@author: Panchdev Singh Chauhan
@email: erpanchdev@gmail.com
"""
import os
import sys

from selenium.webdriver.common.by import By

from nrobo import EnvKeys

# Add host's project path to sys path for module searching...
sys.path.append(os.path.join(os.path.dirname(__file__), ''))

import logging
import os
import sys
from datetime import datetime

import allure
import pytest
import pytest_html
from nrobo.cli import *
from selenium import webdriver
from selenium.webdriver.chrome.service import Service as ChromeService
from webdriver_manager.chrome import ChromeDriverManager

from nrobo.cli.nglobals import *
from nrobo.util.common import *
from nrobo.cli.cli_constants import *
import os.path as path
import nrobo.cli.detection as detect

from nrobo.util.constants import CONST
from nrobo.appium import AUTOMATION_NAMES, CAPABILITY
from selenium.common.exceptions import WebDriverException


def update_pytest_life_cycle_log(life_cycle_item: str, item_type: str = "fixture"):
    if detect.developer_machine():
        from nrobo import NROBO_PATHS
        Common.append_text_to_file(NROBO_PATHS.PYTEST_LIFE_CYCLE_LOGS,
                                   f"\n\n-----------------------\n"
                                   f"Calling from {life_cycle_item} {item_type}")


def update_pytest_life_cycle_log_with_value(value: str):
    if detect.developer_machine():
        from nrobo import NROBO_PATHS
        Common.append_text_to_file(NROBO_PATHS.PYTEST_LIFE_CYCLE_LOGS,
                                   f"\n{value}")


def ensure_logs_dir_exists():
    """checks if driver logs dir exists. if not creates on the fly."""
    from nrobo.cli.cli_constants import NREPORT
    from nrobo import EnvKeys, NROBO_PATHS
    _log_driver_file = NROBO_PATHS.EXEC_DIR / NREPORT.REPORT_DIR / NREPORT.LOG_DIR_DRIVER

    if not _log_driver_file.exists():
        """ensure driver logs dir"""
        try:
            os.makedirs(_log_driver_file)
        except FileExistsError as e:
            pass

    _test_logs_dir = NROBO_PATHS.EXEC_DIR / NREPORT.REPORT_DIR / NREPORT.LOG_DIR_TEST
    if not _test_logs_dir.exists():
        """ensure test logs dir"""
        try:
            os.makedirs(_test_logs_dir)
        except FileExistsError as e:
            pass

    _screenshot_dir = NROBO_PATHS.EXEC_DIR / NREPORT.REPORT_DIR / NREPORT.SCREENSHOTS_DIR
    if not _screenshot_dir.exists():
        """ensure screenshots dir"""
        try:
            os.makedirs(_screenshot_dir)
        except FileExistsError as e:
            pass

    _allure_dir = NROBO_PATHS.EXEC_DIR / NREPORT.ALLURE_REPORT_PATH
    if not _allure_dir.exists():
        """ensure allure dir"""
        try:
            os.makedirs(_allure_dir)
        except FileExistsError as e:
            pass


def read_browser_config_options(_config_path):
    """
    process browser config options from the <_config_path>
    and return list of those.

    If file not found, then raise exception.

    """
    if not _config_path:
        return []

    if path.exists(_config_path):
        """if path exists"""

        # read file and store it's content in a list
        _config_options = []
        with open(_config_path, "r") as f:
            _config_options.append(str(f.readline()).strip())

        return _config_options
    else:
        raise Exception(f"Chrome config file does not exist at path <{_config_path}>!!!")


def add_capabilities_from_file(options):
    """Read capabilities from capability.yaml file

       and add them to the browser options <options>

       and return updated options"""
    from nrobo.util.common import Common
    from nrobo import NROBO_PATHS, Environment, EnvKeys
    if detect.production_machine() and not detect.developer_machine():
        capabilities = Common.read_yaml(NROBO_PATHS.EXEC_DIR / NROBO_PATHS.CAPABILITY_YAML)
    else:
        capabilities = Common.read_yaml(
            NROBO_PATHS.NROBO_DIR / NROBO_PATHS.NROBO / NROBO_PATHS.CAPABILITY_YAML)

    for k, v in capabilities.items():
        options.set_capability(k, v)

    return options


def get_appium_capabilities_from_file():
    """Read appium capabilities from appium_capability.yaml file

       return appium_capabilities"""
    from nrobo.util.common import Common
    from nrobo import NROBO_PATHS, Environment, EnvKeys
    if detect.production_machine() and not detect.developer_machine():
        capabilities = Common.read_yaml(NROBO_PATHS.EXEC_DIR / NROBO_PATHS.CAPABILITY_APPIUM_YAML)
    else:
        capabilities = Common.read_yaml(
            NROBO_PATHS.NROBO_DIR / NROBO_PATHS.NROBO / NROBO_PATHS.CAPABILITY_APPIUM_YAML)

    return capabilities


def pytest_addoption(parser):
    """
    Pass different values to a test function, depending on command line options

    :param parser:
    :return:
    """
    update_pytest_life_cycle_log("pytest_addoption", "hook")

    group = parser.getgroup("nrobo header options")
    # nRoBo appium options
    group.addoption(
        f"--{nCLI.APPIUM}", help=f"Tells nRoBo to trigger via appium client",
        action="store_true", default=False
    )

    # nRoBo webdriver options
    group.addoption(
        f"--{nCLI.BROWSER}", help="""
    Target browser name. Default is chrome.
    Options could be:
        chrome | firefox | safari | edge.
        (Only chrome is supported at present_release.)
    """
    )
    group.addoption(f"--{nCLI.APP}", help="Name of your app project under test")
    group.addoption(f"--{nCLI.REPORT_TITLE}", help="Defines HTML report title")
    group.addoption(f"--{nCLI.URL}", help="Link of application under test")
    group.addoption(f"--{nCLI.USERNAME}", help="Username for login", default="")
    group.addoption(f"--{nCLI.PASSWORD}", help="Password for login", default="")
    group.addoption(f"--{nCLI.BROWSER_CONFIG}", help="Browser config file path for setting requested options")
    group.addoption(f"--{nCLI.PACKAGES}", help="Browser config file path for setting requested options")
    group.addoption(f"--{nCLI.GRID}", help="Url of remote selenium grid server")
    group.addoption(f"--{nCLI.FULLPAGE_SCREENSHOT}",
                    help="Take full page screenshot", action="store_true", default=False)

    # ini option
    parser.addini(f"--{nCLI.APPIUM}", type='bool', help=f"Tells nRoBo to trigger via appium client")
    parser.addini(f"{nCLI.APP}", type="string",
                  help="Name of your app project under test")
    parser.addini(f"{nCLI.REPORT_TITLE}", type="string",
                  help="Defines HTML report title")
    parser.addini(f"{nCLI.URL}", type='string',
                  help="Link of application under test")
    parser.addini(f"{nCLI.USERNAME}", type="string",
                  help="Username for login")
    parser.addini(f"{nCLI.PASSWORD}", type='string',
                  help="Password for login")
    parser.addini(f"{nCLI.BROWSER_CONFIG}", type='string',
                  help="Browser config file path for setting requested options")
    parser.addini(f"{nCLI.BROWSER_CONFIG}", type='string',
                  help="Browser config file path for setting requested options")
    parser.addini(f"{nCLI.PACKAGES}", type='string',
                  help="Browser config file path for setting requested options")
    parser.addini(f"--{nCLI.GRID}", type='string', help="Url of remote selenium grid server")
    parser.addini(f"--{nCLI.FULLPAGE_SCREENSHOT}", type='bool',
                  help="Take full page screenshot")


@pytest.fixture(scope='function')
def url(request):
    """Supply test URL given from nRoBo command line"""
    # Global fixture returning app url
    # Access pytest command line options
    update_pytest_life_cycle_log("url")

    return request.config.getoption(f"--{nCLI.URL}")


@pytest.fixture(scope='function')
def app(request):
    """Supply app name given from nRoBo command line"""

    update_pytest_life_cycle_log("app")

    # Global fixture returning app name
    # Access pytest command line options
    return request.config.getoption(f"--{nCLI.APP}")


@pytest.fixture(scope='function')
def username(request):
    """Supply username given from nRoBo command line"""

    update_pytest_life_cycle_log("username")

    # Global fixture returning admin username
    # Access pytest command line options
    return request.config.getoption(f"--{nCLI.USERNAME}")


@pytest.fixture(scope='function')
def password(request):
    """Supply password given from nRoBo command line"""

    update_pytest_life_cycle_log("password")
    # Global fixture returning admin password
    # Access pytest command line options
    return request.config.getoption(f"--{nCLI.PASSWORD}")


@pytest.fixture(autouse=True, scope='function')
def driver(request):
    """
    Instantiating driver for given browser.
    """

    update_pytest_life_cycle_log("driver")

    # Access pytest command line options
    from nrobo import EnvKeys, console
    browser = request.config.getoption(f"--{nCLI.BROWSER}")

    # get and set url
    _url = request.config.getoption(f"--{nCLI.URL}")
    os.environ[EnvKeys.URL] = _url if _url else CONST.EMPTY

    # get grid url
    _grid_server_url = request.config.getoption(f"--{nCLI.GRID}")

    # initialize driver with None
    _driver = None

    # Set driver log name
    # current test function name
    test_method_name = request.node.name
    from nrobo.cli.cli_constants import NREPORT
    ensure_logs_dir_exists()
    _driver_log_path = NREPORT.REPORT_DIR + os.sep + \
                       NREPORT.LOG_DIR_DRIVER + os.sep + \
                       test_method_name + NREPORT.LOG_EXTENTION

    if int(os.environ[EnvKeys.APPIUM]):

        """get appium driver with given capabilities"""
        from appium import webdriver as _webdriver

        capabilities = get_appium_capabilities_from_file()

        if capabilities[CAPABILITY.AUTOMATION_NAME] == AUTOMATION_NAMES.UI_AUTOMATION2:
            """Create uiautomator2 driver instance"""
            from appium.options.android import UiAutomator2Options

            options = UiAutomator2Options().load_capabilities(capabilities)

            _grid_url_missing = False

            if _grid_server_url is None:
                _grid_url_missing = True
                _grid_server_url = "http://localhost:4723"

            try:
                _driver = _webdriver.Remote(_grid_server_url, options=options)
            except Exception as e:
                if _grid_url_missing:
                    console.rule(f"[{STYLE.HLRed}]\n\nAppium server url is missing![/]\n\n")
                else:
                    console.rule(f"[{STYLE.HLRed}]\n\nIt seems like appium server is not running? "
                                 f"\nor Is appium server url incorrect?"
                                 f"\nPlease check!!![/]\n\n")

    elif browser == Browsers.CHROME:
        """if browser requested is chrome"""

        options = webdriver.ChromeOptions()
        options.add_experimental_option("excludeSwitches", ["enable-automation"])
        options = add_capabilities_from_file(options)

        # enable/disable chrome options from a file
        _browser_options = read_browser_config_options(request.config.getoption(f"--{nCLI.BROWSER_CONFIG}"))
        # apply chrome options
        [options.add_argument(_option) for _option in _browser_options]

        if _grid_server_url:
            """Get instance of remote webdriver"""
            _driver = webdriver.Remote(_grid_server_url,
                                       options=options)
        else:
            """Get instance of local chrom driver"""
            _driver = webdriver.Chrome(options=options,
                                       service=ChromeService(
                                           ChromeDriverManager().install(),
                                           log_output=_driver_log_path))

    elif browser == Browsers.CHROME_HEADLESS:
        """if browser requested is chrome"""

        options = webdriver.ChromeOptions()
        options.add_argument('--headless=new')
        options = add_capabilities_from_file(options)

        # enable/disable chrome options from a file
        _browser_options = read_browser_config_options(
            request.config.getoption(f"--{nCLI.BROWSER_CONFIG}"))
        # apply chrome options
        [options.add_argument(_option) for _option in _browser_options]

        if _grid_server_url:
            """Get instance of remote webdriver"""
            _driver = webdriver.Remote(_grid_server_url,
                                       options=options)
        else:
            """Get instance of local chrom driver"""
            _driver = webdriver.Chrome(options=options,
                                       service=ChromeService(
                                           ChromeDriverManager().install(),
                                           log_output=_driver_log_path))

    elif browser == Browsers.ANTI_BOT_CHROME:
        """if browser requested is anti_bot_chrome"""

        options = webdriver.ChromeOptions()
        options.add_argument('--headless=new')
        options = add_capabilities_from_file(options)

        # enable/disable chrome options from a file
        _browser_options = read_browser_config_options(
            request.config.getoption(f"--{nCLI.BROWSER_CONFIG}"))
        # apply chrome options
        [options.add_argument(_option) for _option in _browser_options]

        if _grid_server_url:
            """Get instance of remote webdriver"""
            from nrobo import console
            console.rule(f"[{STYLE.HLRed}]Anti-Bot Chrome is not supported by Grid Infrastructure![/]")
            sys.exit()
            # _driver = webdriver.Remote(_grid_server_url,
            #                            options=options)
        else:
            """Get instance of local chrom driver"""
            import undetected_chromedriver as uc
            _driver = uc.Chrome(use_subprocess=False, options=options)
            # _driver = webdriver.Chrome(options=options,
            #                            service=ChromeService(
            #                                ChromeDriverManager().install(),
            #                                log_output=_driver_log_path))

    elif browser == Browsers.SAFARI:
        """if browser requested is safari"""

        options = webdriver.SafariOptions()
        options.add_argument("ShowOverlayStatusBar=YES")
        options = add_capabilities_from_file(options)

        # enable/disable chrome options from a file
        _browser_options = read_browser_config_options(
            request.config.getoption(f"--{nCLI.BROWSER_CONFIG}"))
        # apply Safari options
        [options.add_argument(_option) for _option in _browser_options]

        if _grid_server_url:
            """Get instance of remote webdriver"""
            _driver = webdriver.Remote(_grid_server_url,
                                       options=options)
        else:
            """Get instance of local chrom driver"""
            _service = webdriver.SafariService(service_args=["--diagnose"])
            _driver = webdriver.Safari(options=options, service=_service)

    elif browser in [Browsers.FIREFOX, Browsers.FIREFOX_HEADLESS]:
        """if browser requested is firefox"""

        options = webdriver.FirefoxOptions()
        options = add_capabilities_from_file(options)

        if browser == Browsers.FIREFOX_HEADLESS:
            options.add_argument("-headless")

        # enable/disable chrome options from a file
        _browser_options = read_browser_config_options(
            request.config.getoption(f"--{nCLI.BROWSER_CONFIG}"))
        # apply Safari options
        [options.add_argument(_option) for _option in _browser_options]

        if _grid_server_url:
            """Get instance of remote webdriver"""
            _driver = webdriver.Remote(_grid_server_url,
                                       options=options)
        else:
            """Get instance of local firefox driver"""
            _service = webdriver.FirefoxService(log_output=_driver_log_path, service_args=['--log', 'debug'])
            _driver = webdriver.Firefox(options=options, service=_service)

    elif browser == Browsers.EDGE:
        """if browser requested is microsoft edge"""

        options = webdriver.EdgeOptions()
        options = add_capabilities_from_file(options)

        # enable/disable chrome options from a file
        _browser_options = read_browser_config_options(
            request.config.getoption(f"--{nCLI.BROWSER_CONFIG}"))
        # apply Safari options
        [options.add_argument(_option) for _option in _browser_options]

        if _grid_server_url:
            """Get instance of remote webdriver"""
            _driver = webdriver.Remote(_grid_server_url, options=options)
        else:
            """Get instance of local firefox driver"""
            _service = webdriver.EdgeService(log_output=_driver_log_path)
            _driver = webdriver.Edge(options=options, service=_service)

    elif browser == Browsers.IE:
        """if browser requested is microsoft internet explorer"""

        if sys.platform != "win32":
            """No need to proceed"""
            from nrobo.cli.tools import console
            console.rule("IE support available on WIN32 platform only! Quiting test run.")
            console.print(
                """Please note that the Internet Explorer (IE) 11 desktop application ended support 
                for certain operating systems on June 15, 2022. 
                Customers are encouraged to move to Microsoft Edge with IE mode.""")
            exit(1)

        options = webdriver.IeOptions()

        # enable/disable chrome options from a file
        _browser_options = read_browser_config_options(
            request.config.getoption(f"--{nCLI.BROWSER_CONFIG}"))
        # apply Safari options
        [options.add_argument(_option) for _option in _browser_options]

        if _grid_server_url:
            """Get instance of remote webdriver"""
            _driver = webdriver.Remote(_grid_server_url, options=options)
        else:
            """Get instance of local firefox driver"""
            _service = webdriver.IeService(log_output=_driver_log_path)
            _driver = webdriver.Ie(options=options)

    else:
        from nrobo.cli.tools import console
        console.rule(f"[{STYLE.HLRed}]DriverNotConfigured Error!")
        console.print(f"[{STYLE.HLRed}]Driver not configured in nrobo for browser <{browser}>")
        exit(1)

    # store web driver ref in request
    request.node.funcargs['driver'] = _driver
    # yield driver instance to calling test method
    yield _driver

    # quit the browser
    _driver.quit()


@pytest.fixture(scope='function')
def logger(request):
    """
    Instantiate logger instance for each test
    """

    update_pytest_life_cycle_log("logger")

    test_method_name = request.node.name
    from nrobo.cli.cli_constants import NREPORT
    ensure_logs_dir_exists()

    # Setup logger for tests
    logger = logging.getLogger('selenium')
    logger.setLevel(logging.DEBUG)
    _test_logs_path = NREPORT.REPORT_DIR + os.sep + \
                      NREPORT.LOG_DIR_TEST + os.sep + \
                      test_method_name + NREPORT.LOG_EXTENTION
    handler = logging.FileHandler(_test_logs_path)
    logger.addHandler(handler)
    logging.getLogger('selenium.webdriver.remote').setLevel(logging.WARN)
    logging.getLogger('selenium.webdriver.common').setLevel(logging.DEBUG)

    # yield logger instance to calling test method
    yield logger


def pytest_report_header(config):
    """
    Returns console header
    """

    update_pytest_life_cycle_log("pytest_report_header", "hook")

    from nrobo import EnvKeys
    return f"{os.environ[EnvKeys.APP]}" + " test summary".title()


# set up a hook to be able to check if a test has failed
@pytest.hookimpl(tryfirst=True, hookwrapper=True)
def pytest_runtest_makereport(item, call):
    """
    Make report with screenshot attached
    """

    update_pytest_life_cycle_log("pytest_runtest_makereport", "hook")

    outcome = yield
    report = outcome.get_result()

    # test_fn = item.obj
    # docstring = getattr(test_fn, '__doc__')
    # if docstring:
    #     report.nodeid = docstring  # replace __doc__ string with nodeid

    extras = getattr(report, 'extras', [])
    if report.when == 'call':
        xfail = hasattr(report, 'wasxfail')
        if (report.failed and not xfail) \
                or (report.passed and not xfail):
            # Get test method identifier
            node_id = item.nodeid
            # Get reference to test method
            feature_request = item.funcargs['request']
            # Get driver reference from test method by calling driver(request) fixture
            driver = feature_request.getfixturevalue('driver')

            # replace unwanted chars from node id, datetime and prepare a good name for screenshot file
            screenshot_filename = f'{node_id}_{datetime.today().strftime("%Y-%m-%d_%H:%M")}' \
                                  f'_{Common.generate_random_numbers(1000, 9999)}.png' \
                .replace(CONST.FORWARD_SLASH, CONST.UNDERSCORE) \
                .replace(CONST.SCOPE_RESOLUTION_OPERATOR, CONST.UNDERSCORE) \
                .replace(CONST.COLON, CONST.EMPTY).replace('.py', CONST.EMPTY)

            # build screenshot relative path for html report and actual path for saving screenshot
            screenshot_filepath = NREPORT.REPORT_DIR + os.sep + \
                                  NREPORT.SCREENSHOTS_DIR + \
                                  os.sep + screenshot_filename
            screenshot_relative_path = NREPORT.SCREENSHOTS_DIR + os.sep + screenshot_filename

            # Handle fullpagescreenshot cli switch
            fullpagescreenshot = feature_request.config.getoption(f"--{nCLI.FULLPAGE_SCREENSHOT}")

            if fullpagescreenshot:
                driver.maximize_window()
                document_height = \
                    driver.execute_script(
                        'return Math.max( document.body.scrollHeight, document.body.offsetHeight, '
                        'document.documentElement.clientHeight, document.documentElement.scrollHeight, '
                        'document.documentElement.offsetHeight );')
                document_size = driver.get_window_size()
                driver.set_window_size(document_size['width'], document_height)
                driver.find_element(By.TAG_NAME, 'body').screenshot(screenshot_filepath)

            # Attach screenshot to allure report
            try:
                allure.attach(
                    # Not working. Still work in progress...
                    driver.get_screenshot_as_png() if not fullpagescreenshot else driver.find_element(By.TAG_NAME,
                                                                                                      'body').screenshot_as_png,
                    name='screenshot',
                    attachment_type=allure.attachment_type.PNG
                )

                # Attach screenshot to html report
                # Create and save screenshot at <screenshot_filepath>
                if not fullpagescreenshot:
                    driver.save_screenshot(screenshot_filepath)

                # get base64 screenshot
                # failure_screen_shot = driver.get_screenshot_as_base64()

                # attach screenshot with html report. Use relative path to report directory
                extras.append(pytest_html.extras.image(screenshot_relative_path))
                # add relative path to screenshot in the html report
                extras.append(pytest_html.extras.url(screenshot_relative_path))

            except Exception as e:
                extras = []

        # update the report.extras
        report.extras = extras


def pytest_configure(config):
    """
    Description
        configure pytest.
    """
    from nrobo.util.constants import CONST
    update_pytest_life_cycle_log("pytest_configure", "hook")

    os.environ[EnvKeys.TITLE] = str(config.getoption(f'--{nCLI.REPORT_TITLE}')).replace(CONST.UNDERSCORE, CONST.SPACE)
    os.environ[EnvKeys.APP] = str(config.getoption(f'--{nCLI.APP}')).replace(CONST.UNDERSCORE, CONST.SPACE)
    os.environ[EnvKeys.APPIUM] = "1" if str(config.getoption(f'--{nCLI.APPIUM}')) == "True" else "0"

    # add custom markers
    config.addinivalue_line("markers", "sanity: marks as sanity test")
    config.addinivalue_line("markers", "regression: mark as regression test")
    config.addinivalue_line("markers", "ui: mark as ui test")
    config.addinivalue_line("markers", "api: mark as api tests")
    config.addinivalue_line("markers", "nogui: mark as NOGUI tests")
    config.addinivalue_line("markers", "unit: mark as unit test")

    from nrobo import NROBO_PATHS
    if detect.production_machine() and not detect.developer_machine():
        markers = Common.read_yaml(NROBO_PATHS.EXEC_DIR / NROBO_PATHS.MARKERS_YAML)
    else:
        markers = Common.read_yaml(NROBO_PATHS.EXEC_DIR / NROBO_PATHS.NROBO / NROBO_PATHS.MARKERS_YAML)
    for marker, desc in markers.items():
        config.addinivalue_line("markers", f"{marker}: {desc}")


def pytest_metadata(metadata):
    """
    Description
        pytest metadata
    """

    update_pytest_life_cycle_log("pytest_metadata", "hook")

    # pop all the python environment table data
    # metadata.pop("Packages", None)
    # metadata.pop("Platform", None)
    # metadata.pop("Plugins", None)
    # metadata.pop("Python", None)


def pytest_runtest_setup(item):
    from nrobo.util.common import Common
    update_pytest_life_cycle_log("pytest_runtest_setup", "hook")
    from pprint import pprint
    from nrobo.util.common import Common
    try:
        callspec = item.callspec
    except AttributeError as ae:
        callspec = None

    try:
        callobj = item.callobj
    except AttributeError as ae:
        callobj = None

    try:
        fixtureinfo = item.fixtureinfo
    except AttributeError as ae:
        fixtureinfo = None

    update_pytest_life_cycle_log_with_value(f"Function properties:\n"
                                            f"name={item.name}\n"
                                            f"parent={item.parent}\n"
                                            f"config={item.config}\n"
                                            f"callspec={callspec}\n"
                                            f"callobj={callobj}\n"
                                            f"keywords={item.keywords}\n"
                                            f"session={item.session}\n"
                                            f"fixtureinfo={fixtureinfo}\n"
                                            f"originalname={item.originalname}\n"
                                            f"filepath={item.fspath}\n"
                                            f"docstring={item.__doc__}")

    pprint(item.__doc__)
    pprint(item.config)
    print(item.parent)
    pprint(item.name)
    pprint(item.fspath)
    pprint(item.session)
    mod = item.getparent(pytest.Module).obj
    print("Mod.dir")
    pprint(mod.__dir__)
    _class = item.getparent(pytest.Class).obj
    print(f"class where this test {item.name} belong to")
    pprint(_class)
    print(f"All attributes of test class")
    pprint(_class.__dir__(item.name))
    print(f"Get doc string")
    print(_class.__dir__(item.name).__doc__)
    print("Check if item has attribute hello")
    pprint(hasattr(mod, "hello"))
    print(f"node id")
    pprint(item.nodeid)
    print(f"Nodeid dir object")
    pprint(item.nodeid.__dir__)

    if isinstance(item, pytest.Function):
        pass

    # if isinstance(item, pytest.Function):
    #     if not item.fspath.relto(mydir):
    #         return
    #     mod = item.getparent(pytest.Module).obj
    #     if hasattr(mod, "hello"):
    #         print(f"mod.hello {mod.hello!r}")


def pytest_runtest_setup(item):
    # called for running each test in 'a' directory
    update_pytest_life_cycle_log("pytest_runtest_setup", "hook")
    print("setting up", item)


def pytest_html_report_title(report):
    from nrobo import EnvKeys, NROBO_CONST
    from nrobo.cli.cli_constants import NREPORT
    import os
    update_pytest_life_cycle_log("pytest_html_report_title", "hook")

    _suffix = NREPORT.DEFAULT_REPORT_TITLE
    _title_env = os.environ[EnvKeys.TITLE]
    if _title_env not in [_suffix] and _title_env:
        _title = os.environ[EnvKeys.TITLE]
    elif os.environ[EnvKeys.APP].lower() not in [NROBO_CONST.NROBO.lower()]:
        _title = f"{os.environ[EnvKeys.APP]} {_suffix}"
    else:
        _title = f"{_suffix}"

    report.title = _title


def pytest_html_results_summary(prefix, summary, postfix, session):
    """Called before adding the summary section to the report"""

    update_pytest_life_cycle_log("pytest_html_results_summary", "hook")


def pytest_html_results_table_header(cells):
    """Called after building results table header."""

    update_pytest_life_cycle_log("pytest_html_results_table_header", "hook")


def pytest_html_results_table_row(report, cells):
    """Called after building results table row."""

    update_pytest_life_cycle_log("pytest_html_results_table_row", "hook")


def pytest_html_results_table_html(report, data):
    """Called after building results table additional HTML."""

    update_pytest_life_cycle_log("pytest_html_results_table_html", "hook")


def pytest_html_duration_format(duration):
    """Called before using the default duration formatting."""

    update_pytest_life_cycle_log("pytest_html_duration_format", "hook")



"""
=====================CAUTION=======================
DO NOT DELETE THIS FILE SINCE IT IS PART OF NROBO
FRAMEWORK AND IT MAY CHANGE IN THE FUTURE UPGRADES
OF NROBO FRAMEWORK. THUS, TO BE ABLE TO SAFELY UPGRADE
TO LATEST NROBO VERSION, PLEASE DO NOT DELETE THIS
FILE OR ALTER ITS LOCATION OR ALTER ITS CONTENT!!!
===================================================

Definition of nRoBo update utility.

@author: Panchdev Singh Chauhan
@email: erpanchdev@gmail.com
"""
import os
import time
from nrobo.util.network import internet_connectivity
import subprocess
import re
import nrobo.cli.detection as detect


def get_host_version() -> str:
    """get host version of nrobo installation"""
    from nrobo import __version__
    return __version__


def get_pypi_index(package) -> None | str:
    """Get version of <package> from pypi and returns if package is found.
        return None otherwise."""

    if not internet_connectivity():
        """Exit programme."""
        from nrobo import console, STYLE
        console.print(f"[{STYLE.HLRed}]No internet connectivity!")

        # HOW can I proceed without internet connectivity!
        # I need that for performing a few operations.
        # THUS, Quiting NOW!!!
        exit(1)

    result = subprocess.run(['pip', 'index', 'versions', package], text=True, capture_output=True, timeout=10)

    match = re.search(package + r" \(([\d]+[.][\d]+[.][\d]+)\)", result.stdout)

    if match:
        # Tts' a match!
        return match.group(1)
    else:
        # OOPS!, version not found on PyPi. HOW?
        # Don't want to OVERTHINK!
        # Lets' JUST Quit the world of UNKNOWN!
        return None


def update_available() -> bool:
    """Returns version if package is available on pypi
        else returns None otherwise."""
    from nrobo.util.version import Version
    from nrobo import NROBO_CONST
    return Version(get_host_version()) < Version(get_pypi_index(NROBO_CONST.NROBO))


def confirm_update() -> None:
    """Asks host to upgrade.
        Upgrades nrobo if host's reply is affirmative
        else returns with no action"""
    from nrobo.util.version import Version
    from nrobo import NROBO_CONST
    host_version = Version(get_host_version())
    pypi_version = Version(get_pypi_index(NROBO_CONST.NROBO))

    # Apply patches silently
    if Version.present_is_a_patch_release(pypi_version.version, host_version.version) \
            or Version.present_is_a_minor_release(pypi_version.version, host_version.version)\
            or Version.present_is_a_major_release(pypi_version.version, host_version.version):

        from nrobo import console, terminal, STYLE
        from nrobo.cli.ncodes import EXIT_CODES

        version_forced_update = Version(host_version.version)

        if Version.present_is_a_major_release(pypi_version.version, host_version.version):
            request_version = Version.first_major_release(pypi_version.version)
            terminal(['pip', 'install', '--upgrade', f'nrobo=={request_version}', '--require-virtualenv'],
                     debug=False)
            console.rule(f"[{STYLE.HLOrange}]NOTE: A silent major update was installed from "
                         f"{host_version.version} => {pypi_version.version}. Please rerun tests.")
            exit(0)

        elif Version.present_is_a_minor_release(pypi_version.version, host_version.version):

            request_version = Version.first_minor_release(pypi_version.version)
            terminal(['pip', 'install', '--upgrade', f'nrobo=={request_version}', '--require-virtualenv'],
                     debug=False)

            console.rule(f"[{STYLE.HLOrange}]NOTE: A silent minor update was installed from "
                         f"{host_version.version} => {pypi_version.version}. Please rerun tests.")
            exit(0)

        elif host_version <= Version("2024.12.0"):
            # forced update and apply patch delivered in give version

            terminal(['pip', 'install', '--upgrade', f'nrobo=={(host_version + 1).version}', '--require-virtualenv'],
                     debug=False)

        if detect.production_machine() and not detect.developer_machine()\
                and host_version <= Version("2024.12.0"):
            console.rule(f"[{STYLE.HLOrange}]{EXIT_CODES['10001'][1]}")
            exit(EXIT_CODES['10001'][0])

        # return  # Silent patch applied for version 2024.6.10, thus, just return!

    # Ask for major and minor version update
    if host_version < pypi_version:
        # Ok. Since the host version is lower than the latest version,
        # Lets' ask host user if he/she wants to upgrade.
        from nrobo import EnvKeys
        # Enabled Major and Minor release by default.
        # Thus, stopping patch releases from version 2024.25.8
        return

        if int(os.environ[EnvKeys.SUPPRESS_PROMPT]):
            """Return as --suppress switch is supplied"""
            return

        _pypi_version = get_pypi_index(NROBO_CONST.NROBO)
        from nrobo import console, terminal, STYLE
        from rich.prompt import Prompt
        reply = Prompt.ask(
            f"An updated version ({_pypi_version}) is available for nrobo. "
            f"\n Your nRoBo version is {get_host_version()}. "
            f"\n Do you want to upgrade? "
            f"\n(Type [{STYLE.HLGreen}]Yes[/] or [{STYLE.HLRed}]Y[/] to continue. Press any key to skip.)"
            f"\nNOTE: To suppress this propmt, apply CLI switch, --suppress, to your launcher command.")
        if not reply.strip().lower() in ["yes", "y"]:
            # Hmm! Host don't want an update.
            # I don't know why he/she doesn't!!!
            # Anyway, I've had to obey her/his command,
            # Thus, I'm not going to update.
            return  # Bye, Host!

        with console.status("Updating nRoBo"):
            # Host chose to allow update. Lets' update then!

            console.print("Update started")

            return_code = terminal(['pip', 'install', '--upgrade', 'nrobo', '--require-virtualenv'], debug=True)

            if return_code == 0:
                console.print("Update completed successfully.")
            else:
                console.print("Update did not complete.")

            # allow host system to do some stuff!
            time.sleep(2)
